import json

try:
    from django.utils.translation import gettext_lazy
    from django.db.models import JSONField
    from django.forms import JSONField as JSONFormField
# avoid import failures for non Django builds.
except ImportError:
    JSONField = object
    JSONFormField = object
    gettext_lazy = lambda a: a  # noqa: E731


from . import MetaSet


def _recur_serialize_metaset(value):
    """Transform a MetaSet to a JSON serializable value"""
    try:
        return {k: _recur_serialize_metaset(v) for k, v in value.items()}
    except AttributeError:
        return list(value)


class MetaFormField(JSONFormField):
    def prepare_value(self, value):
        if value is None:
            return None
        return super().prepare_value(_recur_serialize_metaset(value))


class MetaSetField(JSONField):
    """A categorized set field."""

    description = gettext_lazy("Dict of sets")

    def __init__(self, *args, **kwargs):
        kwargs["blank"] = True
        kwargs["default"] = MetaSet
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["default"]
        del kwargs["blank"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection, context=None):
        if value is None:
            return value
        if isinstance(value, str):
            # Django >= 3.1
            value = json.loads(value)
        return MetaSet.from_dict(value)

    def to_python(self, value):
        if isinstance(value, MetaSet):
            return value
        if value is None:
            return None
        return MetaSet.from_dict(value)

    def get_db_prep_save(self, value, connection, prepared=False):
        if value is not None:
            value = _recur_serialize_metaset(value)
        return super().get_db_prep_value(value, connection, prepared)

    def validate(self, value, model_instance):
        if value is not None:
            value = _recur_serialize_metaset(value)
        return super().validate(value, model_instance)

    def formfield(self, **kwargs):
        defaults = {"form_class": MetaFormField}
        defaults.update(kwargs)
        return super().formfield(**defaults)



from copy import deepcopy
from itertools import chain, groupby
from importlib.metadata import version


__version__ = version("metaset")


class MetaSet(dict):
    """Basically a dict of set-like objects that behaves as a set.
    Inception is possible.

    Note keys are conserved, even when their content becomes empty.
    They are not conserved in the intersection case,
    if they are not on both ends of the intersection.

    Examples:

    >>> MetaSet(a={1, 2}) & MetaSet(b={3})
    {}
    >>> MetaSet(a={1, 2}) & MetaSet(a={3, 4}, b={1})
    {'a': set()}
    >>> MetaSet(a={1}) - MetaSet(a={1, 2})
    {'a': set()}
    >>> MetaSet(a={1}) ^ MetaSet(a={1, 2})
    {'a': {2}}

    These are less expected:

    >>> MetaSet(a=set()) - MetaSet(a=set())
    {'a': set()}
    >>> MetaSet(a={1}) ^ MetaSet(a={1})
    {'a': set()}
    >>> MetaSet(a=set()) ^ MetaSet(a=set())
    {'a': set()}

    But note symmetric difference has to be permutable:

    >>> MetaSet(a={1}) ^ MetaSet(a={1}) ^ MetaSet(a=set())
    {'a': set()}
    >>> MetaSet(a={1}) ^ MetaSet(a=set()) ^ MetaSet(a={1})
    {'a': set()}

    Finally you can use inception, and this is where key conservation kicks in:

    >>> from pprint import pprint  # for doctest
    >>> pprint(MetaSet(a=MetaSet(b={1, 2}), c=MetaSet(d={3})))
    {'a': {'b': {1, 2}}, 'c': {'d': {3}}}
    >>> MetaSet(a=MetaSet(b={1}), c=MetaSet(d={3})) & MetaSet(a=MetaSet(b={4}))
    {'a': {'b': set()}}
    """

    @classmethod
    def from_dict(cls, value):
        try:
            return cls({k: cls.from_dict(v) for k, v in value.items()})
        except AttributeError:
            return set(value)

    def __isub__(self, rhs):
        for k, v in rhs.items():
            if k in self:
                self[k] -= v
        return self

    def __iand__(self, rhs):
        for k, v in rhs.items():
            if k in self:
                self[k] &= v
        for k in set(self.keys()) - set(rhs.keys()):
            del self[k]
        return self

    def __ixor__(self, rhs):
        for k, v in rhs.items():
            if k in self:
                self[k] ^= v
            else:
                self[k] = v
        return self

    def __ior__(self, rhs):
        for k, v in rhs.items():
            if k in self:
                self[k] |= v
            else:
                self[k] = v
        return self

    def __sub__(self, rhs):
        ret = deepcopy(self)
        ret -= rhs
        return ret

    def __and__(self, rhs):
        ret = deepcopy(self)
        ret &= rhs
        return ret

    def __xor__(self, rhs):
        ret = deepcopy(self)
        ret ^= rhs
        return ret

    def __or__(self, rhs):
        ret = deepcopy(self)
        ret |= rhs
        return ret

    def __lt__(self, rhs):
        return all(k in rhs and v < rhs[k] for k, v in self.items())

    def __le__(self, rhs):
        return all(k in rhs and v <= rhs[k] for k, v in self.items())

    def __gt__(self, rhs):
        return all(k not in rhs or v > rhs[k] for k, v in self.items())

    def __ge__(self, rhs):
        return all(k not in rhs or v >= rhs[k] for k, v in self.items())

    @classmethod
    def union(cls, args):
        """Compute the union of multiple metasets
        Faster than reduce(or_, ...)
        """
        try:
            return cls(
                [
                    (k, cls.union([i[1] for i in g]))
                    for k, g in groupby(
                        sorted(
                            chain.from_iterable(arg.items() for arg in args),
                            key=lambda a: a[0],
                        ),
                        key=lambda a: a[0],
                    )
                ]
            )
        except AttributeError:
            return set(chain.from_iterable(args))
